#!/usr/bin/env python
##   Calverter.py
##
##   Copyright (C) 2007 Mehdi Bayazee (Bayazee@Gmail.com)
##   Copyright (C) 2009 Saeed Rasooli <saeed.gnu@gmail.com> (ilius)
##
##   Iranian (Jalali) calendar: 
##           http://en.wikipedia.org/wiki/Iranian_calendar
##   Islamic (Hijri) calendar:
##           http://en.wikipedia.org/wiki/Islamic_calendar
##   Gregorian calendar:
##           http://en.wikipedia.org/wiki/Gregorian_calendar
##
##   This program is free software; you can redistribute it and/or modify
##   it under the terms of the GNU General Public License as published by
##   the Free Software Foundation; either version 3 of the License,  or
##   (at your option) any later version.
##
##   This program is distributed in the hope that it will be useful, 
##   but WITHOUT ANY WARRANTY; without even the implied warranty of
##   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##   GNU General Public License for more details.
##
##   You should have received a copy of the GNU General Public License along
##   with this program. If not, see <http://www.gnu.org/licenses/gpl.txt>.

from math import floor, ceil
import time, os, sys

def my_raise(File=None):
  i = sys.exc_info()
  if File==None:
    print('line %s: %s: %s'%(i[2].tb_lineno, i[0].__name__, i[1]))
  else:
    print('File %s line %s: %s: %s'%(File, i[2].tb_lineno, i[0].__name__, i[1]))

DATE_JD     = -1
DATE_GREG   = 0
DATE_JALALI = 1
DATE_HIJRI  = 2


minMonthLen={DATE_GREG:29, DATE_JALALI:29, DATE_HIJRI:29}
maxMonthLen={DATE_GREG:31, DATE_JALALI:31, DATE_HIJRI:30}

j_days_in_month = (31, 31, 31, 31, 31, 31, 30, 30, 30, 30, 30, 29)


J0000 = 1721424.5                # Julian date of Gregorian epoch: 0000-01-01
J1970 = 2440587.5                # Julian date at Unix epoch: 1970-01-01
JMJD  = 2400000.5                # Epoch of Modified Julian Date system
J1900 = 2415020.5                # Epoch (day 1) of Excel 1900 date system (PC)
J1904 = 2416480.5                # Epoch (day 0) of Excel 1904 date system (Mac)

## jd ~= time.time()/(24*3600.0) + J1970
## time.time() ~= (jd-J1970)*(24*3600.0) 

#NormLeap = ("Normal year", "Leap year")

GREGORIAN_EPOCH = 1721425.5
HIJRI_EPOCH = 1948439.5
JALALI_EPOCH = 1948320.5


jalaliAlg2820=False
hijriAlg=0
hijriUseDB=True

srcDir = os.path.dirname(sys.argv[0])

hijriDbInitH = (1426,2,1) ; hijriDbInitJD = 2453441 ## load from file
hijriDbEndJD = hijriDbInitJD = 0 ## load from file
hijriMonthLen={} ## load and calc from file

def loadHijriDb(path='%s/hijri.db'%srcDir):
  global hijriDbInitH, hijriMonthLen, hijriDbEndJD, hijriDbInitJD
  if not os.path.isfile(path):
    return
  hijriMonthLenY={}
  lines = file(path).read().split('\n')
  for line in lines[1:]:
    if line!='':
      parts = [int(p) for p in line.split('\t')]
      hijriMonthLenY[parts[0]] = parts[1:]
  parts = [int(p) for p in lines[0].split('\t')]
  hijriDbInitH = parts[:3]
  hijriDbEndJD = hijriDbInitJD = parts[3]
  for y in hijriMonthLenY.keys():
    lst = hijriMonthLenY[y]
    for m  in xrange(len(lst)):
      ml = lst[m]
      if ml!=0:
        hijriMonthLen[y*12+m]=ml
        hijriDbEndJD += ml

loadHijriDb()


def jwday(jd):
   "JWDAY: Calculate day of week from Julian day"
   return int(floor((jd + 1.5))) % 7     

def getWeekDay(y, m, d, mode, firstDay=0):
  return jwday(to_jd(y, m, d, mode)-firstDay)

def weekday_before(weekday, jd):
  """WEEKDAY_BEFORE: Return Julian date of given weekday (0 = Sunday)
                     in the seven days ending on jd."""
  return jd - jwday(jd - weekday)


def search_weekday(weekday, jd, direction, offset):
  """SEARCH_WEEKDAY: Determine the Julian date for: 

                         weekday      Day of week desired, 0 = Sunday
                         jd           Julian date to begin search
                         direction    1 = next weekday, -1 = last weekday
                         offset       Offset from jd to begin search"""
          
  return weekday_before(weekday, jd + (direction * offset))

#  Utility weekday functions, just wrappers for search_weekday

def nearest_weekday(weekday, jd):
  return search_weekday(weekday, jd, 1, 3)

def next_weekday(weekday, jd):
  return search_weekday(weekday, jd, 1, 7)

def next_or_current_weekday(weekday, jd):
  return search_weekday(weekday, jd, 1, 6)

def previous_weekday(weekday, jd):
  return search_weekday(weekday, jd, -1, 1)

def previous_or_current_weekday(weekday, jd):
  return search_weekday(weekday, jd, 1, 0)

def leap_gregorian(year):
  "LEAP_GREGORIAN: Is a given year in the Gregorian calendar a leap year ?"
  return ((year % 4) == 0) and (not(((year % 100) == 0) and ((year % 400) != 0)))

def gregorian_to_jd(year, month, day):
  "GREGORIAN_TO_JD: Determine Julian day number from Gregorian calendar date"
  # Python <= 2.5
  if month <= 2 :
      tm = 0
  elif leap_gregorian(year):
      tm = -1
  else:
      tm = -2
  # Python 2.5
  #tm = 0 if month <= 2 else (-1 if leap_gregorian(year) else -2)
  return (GREGORIAN_EPOCH - 1) + (365 * (year - 1)) + floor((year - 1) / 4) +  (-floor((year - 1) / 100)) + \
         floor((year - 1) / 400) + floor((((367 * month) - 362) / 12) + tm + day)


def jd_to_gregorian(jd) :
  "JD_TO_GREGORIAN: Calculate Gregorian calendar date from Julian day"
  wjd = floor(jd - 0.5) + 0.5
  depoch = wjd - GREGORIAN_EPOCH
  quadricent = floor(depoch / 146097)
  dqc = depoch % 146097
  cent = floor(dqc / 36524)
  dcent = dqc % 36524
  quad = floor(dcent / 1461)
  dquad = dcent % 1461
  yindex = floor(dquad / 365)
  year = int((quadricent * 400) + (cent * 100) + (quad * 4) + yindex)
  if not (cent==4 or yindex==4) :
    year += 1
  yearday = wjd - gregorian_to_jd(year, 1, 1)
  # Python <= 2.5 :
  if wjd < gregorian_to_jd(year, 3, 1):
    leapadj = 0
  elif leap_gregorian(year):
    leapadj = 1
  else:
    leapadj = 2
  # Python 2.5
  #leapadj = 0 if wjd < gregorian_to_jd(year, 3, 1) else (1 if leap_gregorian(year) else 2)
  month = int(floor((((yearday + leapadj) * 12) + 373) / 367))
  day = int(wjd - gregorian_to_jd(year, month, 1)) + 1
  return year, month, day

def n_weeks(weekday, jd, nthweek):
  j = 7 * nthweek
  if nthweek > 0 :
      j += previous_weekday(weekday, jd)
  else :
      j += next_weekday(weekday, jd)
  return j


def iso_to_julian(year, week, day):
   "ISO_TO_JULIAN: Return Julian day of given ISO year, week, and day"
   return day + n_weeks(0, gregorian_to_jd(year - 1, 12, 28), week)


def jd_to_iso(jd):
  "JD_TO_ISO: Return array of ISO (year, week, day) for Julian day"
  year = jd_to_gregorian(jd - 3)[0]
  if jd >= iso_to_julian(year + 1, 1, 1) :
      year += 1
  week = int(floor((jd - iso_to_julian(year, 1, 1)) / 7) + 1)
  day = jwday(jd)
  if day == 0 :
      day = 7
  return year, week, day

def iso_day_to_julian(year, day):
  "ISO_DAY_TO_JULIAN: Return Julian day of given ISO year, and day of year"
  return (day - 1) + gregorian_to_jd(year, 1, 1)

def jd_to_iso_day(jd):
  "JD_TO_ISO_DAY: Return array of ISO (year, day_of_year) for Julian day"
  year = jd_to_gregorian(jd)[0]
  day = int(floor(jd - gregorian_to_jd(year, 1, 1))) + 1
  return year, day

def pad(st, howlong, padwith) :
  "PAD: Pad a string to a given length with a given fill character. "
  s = str(st)
  while s.length < howlong :
      s = padwith + s
  return s


def leap_hijri(year):
  "LEAP_HIJRI: Is a given year a leap year in the Islamic calendar ?"
  return (((year * 11) + 14) % 30) < 11


def hijri_to_jd(year, month, day):
  "HIJRI_TO_JD: Determine Julian day from Islamic date"
  return (day + \
          ceil(29.5 * (month - 1)) + \
          (year - 1) * 354 + \
          floor((3 + (11 * year)) / 30) + \
          HIJRI_EPOCH) - 1

def jd_to_hijri(jd):
  "JD_TO_HIJRI: Calculate Islamic date from Julian day"
  if hijriUseDB:
    jd = int(jd)
    if hijriDbEndJD > jd >= hijriDbInitJD:
      #(yi, mi, di) = hijriDbInitH
      #ymi = yi*12 + mi
      (y, m, d) = hijriDbInitH
      ym = y*12 + m-1
      while jd > hijriDbInitJD:
        monthLen = hijriMonthLen[ym]
        if jd-monthLen > hijriDbInitJD:
          ym += 1
          jd -= monthLen
        elif d+jd-hijriDbInitJD > monthLen:
          ym += 1
          d = d + jd - hijriDbInitJD - monthLen
          jd = hijriDbInitJD
        else:
          d = d + jd - hijriDbInitJD
          jd = hijriDbInitJD
      (y, m) = divmod(ym, 12)
      m+=1
      return (y, m, d)
  jd = floor(jd) + 0.5
  year = int(floor(((30 * (jd - HIJRI_EPOCH)) + 10646) / 10631))
  month = int(min(12, ceil((jd - (29 + hijri_to_jd(year, 1, 1))) / 29.5) + 1))
  day = int(jd - hijri_to_jd(year, month, 1)) + 1
  return year, month, day
 
def jalali_to_jd(year, month, day):
  "JALALI_TO_JD: Determine Julian day from Jalali date"
  if jalaliAlg2820:
    # Python <= 2.5
    if year >=0 :
      rm = 474
    else:
      rm = 473
    epbase = year - (rm)
    # Python 2.5
    #epbase = year - 474 if year>=0 else 473
    epyear = 474 + (epbase % 2820)
    if month <= 7 :
      mm = (month - 1) * 31
    else:
      mm = ((month - 1) * 30) + 6
    return day + mm + \
          floor(((epyear * 682) - 110) / 2816) + \
          (epyear - 1) * 365 + \
          floor(epbase / 2820) * 1029983 + \
          (JALALI_EPOCH - 1)
  else:
    ## Use 33 year algorithm
    ##taken from farsiweb code writen by Roozbeh Pournader <roozbeh@sharif.edu> and Mohammad Toossi <mohammad@bamdad.org> at 2001
    y2 = year-979
    jdays = 365*y2 + (y2/33)*8 + (y2%33+3)/4
    for i in xrange(month-1):
      jdays += j_days_in_month[i]
    jdays += (day-1)
    return int(jdays + 584101 + GREGORIAN_EPOCH) + 0.5

def jd_to_jalali(jd):
  "JD_TO_JALALI: Calculate Jalali date from Julian day"
  if jalaliAlg2820:
    jd = floor(jd) + 0.5
    depoch = jd - jalali_to_jd(475, 1, 1)
    cycle = floor(depoch / 1029983)
    cyear = depoch % 1029983
    if cyear == 1029982 :
      ycycle = 2820
    else :
      aux1 = floor(cyear / 366)
      aux2 = cyear % 366
      ycycle = floor(((2134 * aux1) + (2816 * aux2) + 2815) / 1028522) + aux1 + 1
    year = int(ycycle + (2820 * cycle) + 474)
    if year <= 0 :
      year -= 1
    yday = (jd - jalali_to_jd(year, 1, 1)) + 1
    if yday <= 186:
      month = int(ceil(yday / 31))
    else:
      month = int(ceil((yday - 6) / 30))
    day = int(jd - jalali_to_jd(year, month, 1)) + 1
  else:
    ## Use 33 year algorithm
    ##taken from farsiweb code writen by Roozbeh Pournader <roozbeh@sharif.edu> and Mohammad Toossi <mohammad@bamdad.org> at 2001
    jdays = int(jd - GREGORIAN_EPOCH - 584101)
    # -(1600*365 + 1600/4 - 1600/100 + 1600/400) + 365  -79 +1== -584101
    #print 'jdays =',jdays
    j_np = jdays / 12053
    jdays %= 12053
    year = 979+33*j_np+4*(jdays/1461)
    jdays %= 1461
    if jdays >= 366:
      year += (jdays-1)/365
      jdays = (jdays-1)%365
    month = 12
    for i in xrange(11):
      if jdays >= j_days_in_month[i]:
        jdays -= j_days_in_month[i]
      else:
        month = i+1
        break
    day = jdays+1
  return year, month, day

def leap_jalali(year):
  "LEAP_jalali: Is a given year a leap year in the Jalali calendar ?"
  if jalaliAlg2820:
    if year > 0:
      return (((year-474)%2820 + 512) * 682) % 2816 < 682
    else:
      return (((year-473)%2820 + 512) * 682) % 2816 < 682
  else:
    ## Use 33 year algorithm
    ##taken from farsiweb code writen by Roozbeh Pournader <roozbeh@sharif.edu> and Mohammad Toossi <mohammad@bamdad.org> at 2001
    jy = year - 979
    gdays = ( 365*jy + (jy/33)*8 + (jy%33+3)/4  +  79 ) % 146097
    leap = True
    if gdays >= 36525: # 36525 = 365*100 + 100/4
      gdays -= 1
      gdays = gdays % 36524
      if gdays >= 365:
        gdays += 1
      else:
        leap = False
    if gdays % 1461 >= 366:
      leap = False
    return leap  

def jd_to(jd, target):
  ## hijriAlg:  0:Calvertor,  1:idate,  2:idate_umm_alqura
  if target==DATE_GREG:
    date = jd_to_gregorian(jd)
  elif target==DATE_JALALI:
    date = jd_to_jalali(jd)
  elif target==DATE_HIJRI:
    if hijriAlg==0:
      date = jd_to_hijri(jd)
    elif hijriAlg==1:
      date = jd_to_hijri_idate(jd, umm_alqura=False)
    elif hijriAlg==2:
      date = jd_to_hijri_idate(jd, umm_alqura=True)
    else:
      raise ValueError('invalid hijriAlg=%s passed to function jd_to'%hijriAlg)
  else:
    raise ValueError, 'invalid target date mode: %s'%target
  return date

def to_jd(y, m, d, source):
  if source==DATE_GREG:
    jd = gregorian_to_jd(y, m, d)
  elif source==DATE_JALALI:
    jd = jalali_to_jd(y, m, d)
  elif source==DATE_HIJRI:
    jd = hijri_to_jd(y, m, d)
  else:
    raise ValueError, 'invalid source date mode: %s'%source
  return jd

def convert(y, m, d, source=DATE_GREG, target=DATE_JALALI):
  return jd_to(to_jd(y, m, d, source), target)


def jd_to_hijri_idate(jd, umm_alqura=False):
  import os
  (y, m, d) = jd_to_gregorian(jd)
  fixed = str(y).zfill(4) + str(m).zfill(2) + str(d).zfill(2)
  cmd = 'idate --gregorian %s'%fixed
  if umm_alqura:
    cmd += ' --umm_alqura'
  output = os.popen(cmd).read()
  try:
    (hd, hm, hy) = output.split('\n')[3].split(':')[1].split('/')
    hy = hy.split(' ')[0]
  except:
    my_raise(__file__)
    raise OSError('command "idate" not found')
  return (int(hy), int(hm), int(hd))


########################################################
def gmtime(sec=None, mode=DATE_GREG):
  if mode==DATE_GREG:
    return time.gmtime(sec)
  t = time.gmtime(sec)
  t[:3] = convert(t[0], t[1], t[2], DATE_GREG, mode)
  return t

def localtime(sec=None, mode=DATE_GREG):
  if mode==DATE_GREG:
    return list(time.localtime(sec))
  t = list(time.localtime(sec))
  t[:3] = convert(t[0], t[1], t[2], DATE_GREG, mode)
  return t

def mktime(t, mode=DATE_GREG):
  if mode==DATE_GREG:
    return time.mktime(t)
  t[:3] = convert(t[0], t[1], t[2], mode, DATE_GREG)
  return time.mktime(t)





#########################################################
def test_hijri_with_idate(self):
  jd_today = int(gregorian_to_jd(*time.localtime()[:3]))
  for jd in xrange(jd_today-400, jd_today+10):
    greg   = jd_to_gregorian(jd)
    hijri  = jd_to_hijri(jd)
    hijri2 = jd_to_hijri_idate(jd)
    if hijri != hijri2:
      print greg, hijri, hijri2

def test_jalali_with_farsiweb(self):
  from utility_disable import jalali_to_milady
  jd_start = int(to_jd(1305, 1, 1, DATE_JALALI))
  jd_end   = int(to_jd(1500,12,29, DATE_JALALI))
  for j in xrange(jd_start, jd_end+1):
    jd = j + 0.5
    try:
      jdate1 = jd_to_jalali(jd)
      gdate  = jd_to_gregorian(jd) ### int+0.5
      jd2 = jalali_to_jd(*jdate1)
      if jd!=jd2:		### OK
        print jd, jd2
        sys.exit(1)
      jdate2 = convert(gdate[0], gdate[1], gdate[2], DATE_GREG, DATE_JALALI)
      if jdate1 != jdate2:	### OK
        print 'jdate1 =', jdate1, '\tjdate2 =', jdate2
        sys.exit(1)
      gdate2   = convert(jdate2[0], jdate2[1], jdate2[2], DATE_JALALI, DATE_GREG)
      if gdate != gdate2:	### OK
        print 'gdate =', gdate, '\tgdate2 =', gdate2
        sys.exit(1)
      #### Testing with farsiweb algorithm
      out = os.popen("/data/Data-New/Programming/Calendar-work/StarCalendar/jalali %s %s %s"%gdate).read().split('/')
      jdate3 = (int(out[0]), int(out[1]), int(out[2]))
      if jdate1 != jdate3:### Breaks in 2025/3/20:	jdate1 = (1404, 1, 1) 	jdate3 = (1403, 12, 30)
        print 'gdate =', gdate, '\tjdate1 =', jdate1, '\tjdate3 =', jdate3
        sys.exit(1)
    except KeyboardInterrupt:
      print 'Stoped!'
      try:
        raw_input()
      except KeyboardInterrupt:
        sys.exit(0)



if __name__=='__main__':
  import sys, os
  c = Calverter()
  arg = sys.argv
  """
  #c.test_hijri_with_idate()
  c.test_jalali_with_farsiweb()
  print 'Test was succesfull'
  #sp = arg[1].split('/')
  #print c.gregorian_to_jd(int(sp[0]), int(sp[1]), int(sp[2]))
  sys.exit(0)
  
  print hijriDbInitDates
  print
  for i in (0,1,2):
    (y,m,d) = hijriDbInitDates[i]
    jd = c.to_jd(y, m, d, i)
    print jd, c.jd_to_hijri(jd)
  print
  for i in (0, 1, 2):
    print c.jd_to(hijriDbInitJD, i)
  
  jdi = hijriDbInitJD
  for jd in xrange(jdi, jdi+1000):
    date1 = c.jd_to_hijri(jd)
    date2 = c.jd_to_hijri(jd)
    if date1!=date2:
      print date1, date2
  """
  if len(arg)<3:
    sys.exit(0)
  if arg[1]=='-j':
    mode=DATE_JALALI
  elif arg[1]=='-g':
    mode=DATE_GREG
  elif arg[1] in ('-h', '-i'):
    mode=DATE_HIJRI
  elif arg[1]=='-jd':
    mode=DATE_JD
  else:
    sys.exit(1)
  if mode==DATE_JD:
    jd = int(arg[2])
  else:
    sp = arg[2].split('/')
    y = int(sp[0])
    m = int(sp[1])
    d = int(sp[2])
    jd = c.to_jd(y, m, d, mode)
  print 'Julian Day: %s'%jd
  print 'Gregorian:  '+('%2d/%2d/%2d'%c.jd_to(jd, DATE_GREG)).replace(' ', '0')
  print 'Jalali:     '+('%2d/%2d/%2d'%c.jd_to(jd, DATE_JALALI)).replace(' ', '0')
  print 'Hijri:      '+('%2d/%2d/%2d'%c.jd_to(jd, DATE_HIJRI)).replace(' ', '0')
      
  
  
  

