#!/usr/bin/python
"""This module provides features to help with debugging ROX applications."""

import sys, os, gtk
import traceback
import gobject
import linecache


_ = str

def set_translator(translator):
  global _
  _ = translator


class ButtonMixed(gtk.Button):
	"""A button with a standard stock icon, but any label. This is useful
	when you want to express a concept similar to one of the stock ones."""
	def __init__(self, stock, message):
		"""Specify the icon and text for the new button. The text
		may specify the mnemonic for the widget by putting a _ before
		the letter, eg:
		button = ButtonMixed(gtk.STOCK_DELETE, '_Delete message')."""
		gtk.Button.__init__(self)
	
		label = gtk.Label('')
		label.set_text_with_mnemonic(message)
		label.set_mnemonic_widget(self)

		image = gtk.image_new_from_stock(stock, gtk.ICON_SIZE_BUTTON)
		box = gtk.HBox(False, 2)
		align = gtk.Alignment(0.5, 0.5, 0.0, 0.0)

		box.pack_start(image, False, False, 0)
		box.pack_end(label, False, False, 0)

		self.add(align)
		align.add(box)
		align.show_all()



def info(message):
	"Display informational message. Returns when the user closes the box."
	#toplevel_ref()
	box = gtk.MessageDialog(None, 0, gtk.MESSAGE_INFO, gtk.BUTTONS_OK, message)
	box.set_position(gtk.WIN_POS_CENTER)
	box.set_title(_('Information'))
	box.run()
	box.destroy()
	#toplevel_unref()


def alert(message):
	"Display message in an error box. Return when the user closes the box."
	#toplevel_ref()
	box = gtk.MessageDialog(None, 0, gtk.MESSAGE_ERROR, gtk.BUTTONS_OK, message)
	box.set_position(gtk.WIN_POS_CENTER)
	box.set_title(_('Error'))
	box.run()
	box.destroy()
	#toplevel_unref()


#from rox import toplevel_ref, toplevel_unref ##??????????
#from rox import alert

savebox = None

def _show_debug_help():
	from os.path import join, dirname
	help = join(dirname(dirname(dirname(__file__))), 'Help', 'Errors')
	from rox import filer
	filer.spawn_rox((help,))

def show_exception(type, value, tb, auto_details = False):
	"""Display this exception in an error box. The user has the options
	of ignoring the error, quitting the application and examining the
	exception in more detail. See also rox.report_exception()."""

	QUIT = 1
	DETAILS = 2
	SAVE = 3
	
	brief = ''.join(traceback.format_exception_only(type, value))

	#toplevel_ref()
	box = gtk.MessageDialog(None, 0, gtk.MESSAGE_ERROR, gtk.BUTTONS_NONE, brief)
	
	if not auto_details:
		button = ButtonMixed(gtk.STOCK_ZOOM_IN, _('_Details'))
		button.set_flags(gtk.CAN_DEFAULT)
		button.show()
		box.add_action_widget(button, DETAILS)

	box.add_button(gtk.STOCK_HELP, gtk.RESPONSE_HELP)
	box.add_button(gtk.STOCK_OK, gtk.RESPONSE_OK)
	box.set_default_response(gtk.RESPONSE_OK)

	box.set_position(gtk.WIN_POS_CENTER)
	box.set_title(_('Error'))
	reply = []
	def response(box, resp):
		reply.append(resp)
		gtk.main_quit()
	box.connect('response', response)
	box.show()

	if tb:
		bug_report = 'Traceback (most recent call last):\n' + \
			     ''.join(traceback.format_stack(tb.tb_frame.f_back) +
				     traceback.format_tb(tb) +
				     traceback.format_exception_only(type, value))
	else:
		bug_report = 'No stack trace.'

	while 1:
		if auto_details:
			resp = DETAILS
			auto_details = False
		else:
			gtk.main()
			resp = reply.pop()
		if resp == int(gtk.RESPONSE_OK) or resp == int(gtk.RESPONSE_DELETE_EVENT):
			break
		if resp == SAVE:
			global savebox
			if savebox:
				savebox.destroy()
			def destroy(box):
				global savebox	# For pychecker
				savebox = None
			#from saving import StringSaver##??????????
			savebox = StringSaver(bug_report, 'BugReport')
			savebox.connect('destroy', destroy)
			savebox.show()
			continue
		if resp == QUIT:
			sys.exit(1)
		elif resp == int(gtk.RESPONSE_HELP):
			_show_debug_help()
			continue
		assert resp == DETAILS
		box.set_response_sensitive(DETAILS, False)

		button = ButtonMixed(gtk.STOCK_SAVE, _('_Bug Report'))
		button.set_flags(gtk.CAN_DEFAULT)
		button.show()
		box.add_action_widget(button, SAVE)
		box.action_area.set_child_secondary(button, True)

		button = ButtonMixed(gtk.STOCK_QUIT, _('Forced _Quit'))
		button.set_flags(gtk.CAN_DEFAULT)
		button.show()
		box.add_action_widget(button, QUIT)
		box.action_area.set_child_secondary(button, True)

		if tb:
			ee = ExceptionExplorer(tb)
			box.vbox.pack_start(ee)
			ee.show()
		else:
			no_trace = gtk.Label('No traceback object!')
			box.vbox.pack_start(no_trace)
			no_trace.show()
	box.destroy()
	#toplevel_unref()

class ExceptionExplorer(gtk.Frame):
	"""Displays details from a traceback object."""
	LEAF = 0
	LINE = 1
	FUNC = 2
	CODE = 3
	FILE = 4
	def __init__(self, tb):
		gtk.Frame.__init__(self, _('Stack trace (innermost last)'))

		vbox = gtk.VBox(False, 0)
		self.add(vbox)

		inner = gtk.Frame()
		inner.set_shadow_type(gtk.SHADOW_IN)
		vbox.pack_start(inner, False, True, 0)

		self.savebox = None

		self.tb = tb
		
		self.model = gtk.ListStore(gobject.TYPE_STRING, gobject.TYPE_INT,
					 gobject.TYPE_STRING, gobject.TYPE_STRING,
					 gobject.TYPE_STRING)
		tree = gtk.TreeView(self.model)
		inner.add(tree)

		cell = gtk.CellRendererText()

		column = gtk.TreeViewColumn('File', cell, text = ExceptionExplorer.LEAF)
		cell.set_property('xalign', 1)
		tree.append_column(column)

		cell = gtk.CellRendererText()
		column = gtk.TreeViewColumn('Line', cell, text = ExceptionExplorer.LINE)
		tree.append_column(column)
		column = gtk.TreeViewColumn('Func', cell, text = ExceptionExplorer.FUNC)
		tree.append_column(column)
		column = gtk.TreeViewColumn('Code', cell, text = ExceptionExplorer.CODE)
		tree.append_column(column)

		inner.set_border_width(5)

		frames = []
		while tb is not None:
			frames.insert(0, (tb.tb_frame, traceback.tb_lineno(tb)))
			tb = tb.tb_next
		f = self.tb.tb_frame
		if f:
			f = f.f_back	# Skip the reporting frame
		while f is not None:
			frames.append((f, f.f_lineno))
			f = f.f_back

		frames.reverse()

		new = None
		for f, lineno in frames:
			co = f.f_code
			filename = co.co_filename
			name = co.co_name
			line = linecache.getline(filename, lineno).strip()

			leafname = os.path.basename(filename)
			
			new = self.model.append()
			self.model.set(new, ExceptionExplorer.LEAF, leafname,
					    ExceptionExplorer.LINE, lineno,
					    ExceptionExplorer.FUNC, name,
					    ExceptionExplorer.CODE, line,
					    ExceptionExplorer.FILE, filename)

		def selected_frame():
			selected = sel.get_selected()
			assert selected
			model, titer = selected
			frame, = model.get_path(titer)
			return frames[frame][0]

		vars = gtk.ListStore(str, str)
		sel = tree.get_selection()
		sel.set_mode(gtk.SELECTION_BROWSE)
		def select_frame(tree):
			vars.clear()
			for n, v in selected_frame().f_locals.iteritems():
				value = `v`
				if len(value) > 500:
					value = value[:500] + ' ...'
				new = vars.append()
				vars.set(new, 0, str(n), 1, value)
		sel.connect('changed', select_frame)
		def show_source(tree, path, column):
			line = self.model[path][ExceptionExplorer.LINE]
			file = self.model[path][ExceptionExplorer.FILE]
			import launch
			launch.launch('http://rox.sourceforge.net/2005/interfaces/Edit',
					'-l%d' % line, file)
			
		tree.connect('row-activated', show_source)

		# Area to show the local variables
		tree = gtk.TreeView(vars)

		vbox.pack_start(gtk.Label(_('Local variables in selected frame:')),
				False, True, 0)

		cell = gtk.CellRendererText()
		column = gtk.TreeViewColumn('Name', cell, text = 0)
		cell.set_property('xalign', 1)
		tree.append_column(column)
		cell = gtk.CellRendererText()
		column = gtk.TreeViewColumn('Value', cell, text = 1)
		tree.append_column(column)

		inner = gtk.ScrolledWindow()
		inner.set_size_request(-1, 200)
		inner.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_ALWAYS)
		inner.set_shadow_type(gtk.SHADOW_IN)
		inner.add(tree)
		inner.set_border_width(5)
		vbox.pack_start(inner, True, True, 0)

		if new:
			sel.select_iter(new)

		hbox = gtk.HBox(False, 4)
		hbox.set_border_width(5)
		vbox.pack_start(hbox, False, True, 0)
		hbox.pack_start(gtk.Label('>>>'), False, True, 0)

		expr = gtk.Entry()
		hbox.pack_start(expr, True, True, 0)
		def activate(entry):
			expr = entry.get_text()
			frame = selected_frame()
			try:
				info(`eval(expr, frame.f_locals, frame.f_globals)`)
			except:
				extype, value = sys.exc_info()[:2]
				brief = ''.join(traceback.format_exception_only(extype, value))
				alert(brief)
			entry.grab_focus()
		expr.connect('activate', activate)

		vbox.show_all()
